# Ngaro VM
# Copyright (c) 2010 - 2011, Charles Childers
# ----------------------------------------------------------------------------- Dependencies
use strict;
use POSIX;
use integer;

package Retro;
local $| = 1;
# ----------------------------------------------------------------------------- Variables
our $data    = 0;
our $i       = 0;
our $n       = 0;
our $sp      = 0;
our $rp      = 0;
our $ip      = 0;
our @memory  = ();
our @stack   = ();
our @address = ();
our @ports   = ();
# ----------------------------------------------------------------------------- Save Image
sub save_image {
  open f, ">retroImage1" or die #!;
  binmode f;
  my $i = 0;
  while ($i < @memory[3]) {
    print f pack("i", @memory[$i]);
    $i++;
  }
  close f;
}
# ----------------------------------------------------------------------------- I/O Handler
sub handle_devices {
  if (@ports[0] == 0)
  {
    @ports[0] = 1;
    if (@ports[1] == 1)
    {
      $i = getc(STDIN);
      @ports[1] = ord($i);
      if (@ports[1] == 13)
      {
        @ports[1] = 10;
      }
    }
    if (@ports[2] == 1)
    {
      if (@stack[$sp] < 0)
      {
        print "...";
      }
      else
      {
        print chr(@stack[$sp]);
      }
      $sp = $sp - 1;
      @ports[2] = 0;
    }

    if (@ports[4] == 1)
    {
      save_image();
      @ports[4] = 0;
    }


    if (@ports[5] == -1)
    {
      @ports[5] = 1000000;
    }
    if (@ports[5] == -2)
    {
      @ports[5] = 0;
    }
    if (@ports[5] == -3)
    {
      @ports[5] = 0;
    }
    if (@ports[5] == -4)
    {
      @ports[5] = 0;
    }
    if (@ports[5] == -5)
    {
      @ports[5] = $sp;
    }
    if (@ports[5] == -6)
    {
      @ports[5] = $rp;
    }
    if (@ports[5] == -7)
    {
      @ports[5] = 0;
    }
    if (@ports[5] == -8)
    {

      @ports[5] = time;
    }
    if (@ports[5] == -9)
    {
      $ip = 1000000;
      @ports[5] = 0;
    }
    if (@ports[5] == -10)
    {
      @ports[5] = 0;
    }
    if (@ports[5] == -11)
    {
      @ports[5] = 0;
    }
    if (@ports[5] == -12)
    {
      @ports[5] = 0;
    }
  }
}
# ----------------------------------------------------------------------------- Process Opcodes
sub process {
  my $opcode = @memory[$ip];
  my $a = 0;
  my $b = 0;
  my $c = 0;
  my $d = 0;
  if    ($opcode == 0)  # nop
  {
  }
  elsif ($opcode == 1)  # lit
  {
    $sp++; $ip++;
    @stack[$sp] = @memory[$ip];
  }
  elsif ($opcode == 2)  # dup
  {
    $sp++;
    @stack[$sp] = @stack[$sp - 1];
  }
  elsif ($opcode == 3)  # drop
  {
    $sp--;
  }
  elsif ($opcode == 4)  # swap
  {
    $a = @stack[$sp];
    @stack[$sp] = @stack[$sp - 1];
    @stack[$sp - 1] = $a;
  }
  elsif ($opcode == 5)  # push
  {
    $rp++;
    @address[$rp] = @stack[$sp];
    $sp--;
  }
  elsif ($opcode == 6)  # pop
  {
    $sp++;
    @stack[$sp] = @address[$rp];
    $rp--;
  }
  elsif ($opcode == 7)  # loop
  {
    @stack[$sp]--;
    if (@stack[$sp] != 0 && @stack[$sp] > -1)
    {
      $ip++;
      $ip = @memory[$ip] - 1;
    }
    else
    {
      $ip++;
      $sp--;
    }
  }
  elsif ($opcode == 8)  # jump
  {
    $ip++;
    $ip = @memory[$ip] - 1;
  }
  elsif ($opcode == 9)  # return
  {
    $ip = @address[$rp];
    $rp--;
  }
  elsif ($opcode == 10) # > jump
  {
    $ip++;
    if (@stack[$sp - 1] > @stack[$sp]) { $ip = @memory[$ip] - 1; }
    $sp = $sp - 2;
  }
  elsif ($opcode == 11) # < jump
  {
    $ip++;
    if (@stack[$sp - 1] < @stack[$sp]) { $ip = @memory[$ip] - 1; }
    $sp = $sp - 2;
  }
  elsif ($opcode == 12) # != jump
  {
    $ip++;
    if (@stack[$sp - 1] != @stack[$sp]) { $ip = @memory[$ip] - 1; }
    $sp = $sp - 2;
  }
  elsif ($opcode == 13) # == jump
  {
    $ip++;
    if (@stack[$sp - 1] == @stack[$sp]) { $ip = @memory[$ip] - 1; }
    $sp = $sp - 2;
  }
  elsif ($opcode == 14) # @
  {
    @stack[$sp] = @memory[@stack[$sp]];
  }
  elsif ($opcode == 15) # !
  {
    @memory[@stack[$sp]] = @stack[$sp - 1];
    $sp = $sp - 2;
  }
  elsif ($opcode == 16) # +
  {
    @stack[$sp - 1] = @stack[$sp] + @stack[$sp - 1];
    $sp--;
  }
  elsif ($opcode == 17) # -
  {
    @stack[$sp - 1] = @stack[$sp - 1] - @stack[$sp];
    $sp--;
  }
  elsif ($opcode == 18) # *
  {
    @stack[$sp - 1] = @stack[$sp - 1] * @stack[$sp];
    $sp--;
  }
  elsif ($opcode == 19) # /mod
  {
    $a = @stack[$sp];
    $b = @stack[$sp - 1];
    @stack[$sp] = POSIX::floor($b / $a);
    @stack[$sp - 1] = $b % $a;
  }
  elsif ($opcode == 20) # and
  {
    $a = @stack[$sp];
    $b = @stack[$sp - 1];
    $sp--;
    @stack[$sp] = $b & $a;
  }
  elsif ($opcode == 21) # or
  {
    $a = @stack[$sp];
    $b = @stack[$sp - 1];
    $sp--;
    @stack[$sp] = $b | $a;
  }
  elsif ($opcode == 22) # xor
  {
    $a = @stack[$sp];
    $b = @stack[$sp - 1];
    $sp--;
    @stack[$sp] = $b ^ $a;
  }
  elsif ($opcode == 23) # <<
  {
    $a = @stack[$sp];
    $b = @stack[$sp - 1];
    $sp--;
    @stack[$sp] = $b << $a;
  }
  elsif ($opcode == 24) # >>
  {
    $a = @stack[$sp];
    $b = @stack[$sp - 1];
    $sp--;
    @stack[$sp] = $b >> $a;
  }
  elsif ($opcode == 25) # 0;
  {
    if (@stack[$sp] == 0)
    {
      $sp--;
      $ip = @address[$rp];
      $rp--;
    }
  }
  elsif ($opcode == 26) # inc
  {
    @stack[$sp]++;
  }
  elsif ($opcode == 27) # dec
  {
    @stack[$sp]--;
  }
  elsif ($opcode == 28) # in
  {
    $a = @stack[$sp];
    @stack[$sp] = @ports[$a];
    @ports[$a] = 0;
  }
  elsif ($opcode == 29) # out
  {
    @ports[@stack[$sp]] = @stack[$sp - 1];
    $sp = $sp - 2;
  }
  elsif ($opcode == 30) # wait
  {
    handle_devices();
  }
  else                  # call (implicit)
  {
    $rp++;
    @address[$rp] = $ip;
    $ip = @memory[$ip] - 1;
  }
}
# ----------------------------------------------------------------------------- Load Image
open f, "retroImage1" or die #!;
binmode f;
$i = 0;
while (($n = read(f, $data, 4)) != 0) {
  @memory[$i] = int(unpack("i", $data));
  $i++;
}
close f;
# ----------------------------------------------------------------------------- Initialize
while ($i < 1000000) {
  @memory[$i] = 0;
  $i++;
}
$i = 0;
while ($i < 1024) {
  @stack[$i] = 0;
  @address[$i] = 0;
  @ports[$i] = 0;
  $i++;
}
# ----------------------------------------------------------------------------- main loop
while ($ip < 1000000)
{
  process();
  $ip++;
}
